# fna_ro.tcl:  Implement read-only tag for text widgets.
# RCS Id: $Id: fna_ro.tcl,v 1.0.0.1 1996/05/24 06:09:15 ndanger Alpha $

# fna:ro_version:  Return the version of this file.
proc fna:ro_version {} {
  return [fna:ver_rcs {$Revision: 1.0.0.1 $} {$State: Alpha $}]
}


# fna_ro.tcl:  implement a read-only tag for text widgets.  
# Invocation:
#	fna:ro
# Description:
# 	Characters tagged with the tag name 'readonly' will not be editable.
# 	For the most part this is accomplished by modifying the routine
# 	tkTextInsert to check for the tag before inserting any characters.
# 	Bindings that result in deletion of characters have to be handled
# 	separately, so these bindings are redefined.  Most of this code
#	is copied directly out of text.tcl.  Note that for tk3.6, we need
#	to force reading of text.tcl so that if this proc is called before
#	a text widget is created, the default class bindings don't wipe
#	these ones out.  This isn't a problem for tk4.x, because later
#	versions ensure that class bindings are read in early on.

proc fna:ro {} {
  
  global tk_version
  global tk_strictMotif
  
  if {$tk_version >= 4.0} {
  
    bind Text <Delete> {
      if {[%W tag nextrange sel 1.0 end] != ""} {
        if {[fna_int:ro_readonlytag %W sel.first sel.last]} \
          {return} {%W delete sel.first sel.last}
      } else {
        if {[fna_int:ro_readonlytag %W insert]} {return}
        %W delete insert
        %W see insert
      }
    }
    bind Text <BackSpace> {
      if {[%W tag nextrange sel 1.0 end] != ""} {
        if {[fna_int:ro_readonlytag %W sel.first sel.last]} \
          {return} {%W delete sel.first sel.last}
      } else {
        if {[fna_int:ro_readonlytag %W insert-1c]} {return}
        %W delete insert-1c
        %W see insert
      }
    }
    
    # Additional emacs-like bindings:
    
    if !$tk_strictMotif {
        bind Text <Control-d> {
    	if {![fna_int:ro_readonlytag %W insert]} {%W delete insert}
        }
        bind Text <Control-k> {
    	if [%W compare insert == {insert lineend}] {
    	    if {![fna_int:ro_readonlytag %W insert]} {%W delete insert}
    	} else {
    	    %W delete insert [fna_int:ro_nextreadonly %W insert]
    	}
        }
        bind Text <Control-o> {
          if {![fna_int:ro_readonlytag %W insert]} {
    	%W insert insert \n
    	%W mark set insert insert-1c
          }
        }
        bind Text <Control-t> {
    	tkTextTranspose %W
        }
        bind Text <Meta-d> {
            if {[fna_int:ro_readonlytag %W insert {insert wordend}]} {
              %W delete insert [fna_int:ro_nextreadonly %W insert]
            } else {
    	  %W delete insert {insert wordend}
    	}
        }
        bind Text <Meta-BackSpace> {
          	if {[fna_int:ro_readonlytag %W {insert -1c wordstart} insert]} {
              %W delete [fna_int:ro_prevreadonly %W {insert-1c}]
            } else {
              %W delete {insert -1c wordstart} insert
            }
        }
    
        # A few additional bindings of my own.
    
        bind Text <Control-h> {
    	if {([%W compare insert != 1.0]) && \
                (![fna_int:ro_readonlytag %W insert-1c])} {
    	    %W delete insert-1c
    	    %W see insert
    	}
        }
    }
    
    # tkTextInsert --
    # Insert a string into a text at the point of the insertion cursor.
    # If there is a selection in the text, and it covers the point of the
    # insertion cursor, then delete the selection before inserting.
    #
    # Arguments:
    # w -		The text window in which to insert the string
    # s -		The string to insert (usually just a single character)
    
    proc tkTextInsert {w s} {
        if {($s == "") || ([$w cget -state] == "disabled")} {
    	return
        }
        if {[fna_int:ro_readonlytag $w insert] && [fna_int:ro_readonlytag $w insert-1c] && \
            ([$w index insert] != 1.0)} {return}
        catch {
    	if {[$w compare sel.first <= insert]
    		&& [$w compare sel.last >= insert]} {
    	    $w delete sel.first sel.last
    	}
        }
        $w insert insert $s
        $w see insert
    }
    
  
  } else {
    
    source $tk_library/tk.tcl
    source $tk_library/text.tcl
    
    proc tk_textBackspace w {
        if {[fna_int:ro_readonlytag $w insert-1c]} {return}
        $w delete insert-1c insert
    }
  
    bind Text <Any-KeyPress> {
      if {[fna_int:ro_readonlytag %W insert] && [fna_int:ro_readonlytag %W insert-1c] && \
          ([%W index insert] != 1.0) && ([%W compare [%W index insert] == end])} \
          {return}
      if {"%A" != ""} {
  	%W insert insert %A
  	if {[fna_int:ro_readonlytag %W insert-1c]} {%W tag remove readonly insert-1c}
  	%W yview -pickplace insert
      }
    }
    bind Text <Return> {
      if {[fna_int:ro_readonlytag %W insert]} return
      %W insert insert \n; %W yview -pickplace insert
    }
    bind Text <Control-d> {
      if {[fna_int:ro_readonlytag %W sel.first sel.last]} \
        {return} {%W delete sel.first sel.last}
    }
    bind Text <Control-v> {
        if {[fna_int:ro_readonlytag %W insert]} return
        %W insert insert [selection get]
        %W yview -pickplace insert
    }
  
  }
}


proc fna_int:ro_readonlytag {w i1 {i2 ""}} {
  
  if {$i2 == ""} {
    if {[lsearch -exact [$w tag names $i1] readonly] != -1} \
      {return 1} {return 0}
  } else {
    for {set i 0} {[$w compare "${i1}+${i}c" < $i2]} {incr i} {
      if {[lsearch -exact [$w tag names $i1+${i}c] readonly] != -1} {return 1}
    }
    return 0
  }
  
}

proc fna_int:ro_nextreadonly {w i} {
  
  if {![fna_int:ro_readonlytag $w $i "$i lineend"]} {return "$i lineend"}
  for {set j 0} {[$w compare "${i}+${j}c" <= "${i} lineend"]} {incr j} {
    if {[fna_int:ro_readonlytag $w "${i}+${j}c"]} {
      return [$w index ${i}+${j}c]
    }
  }
  
}

proc fna_int:ro_prevreadonly {w i} {
  
  if {![fna_int:ro_readonlytag $w "$i linestart" $i]} {return "$i linestart"}
  for {set j 0} {[$w compare "${i}-${j}c" >= "${i} linestart"]} {incr j} {
    if {[fna_int:ro_readonlytag $w "${i}-${j}c"]} {
      return [$w index ${i}-${j}c]
    }
  }
  
}
